@@ -277,3 +277,115 @@ class TestOrder(unittest.TestCase):
         self.assertEqual(g3, [{'col1': 'c', 'col2': 7}, {
                          'col1': 'c', 'col2': 4}, {'col1': 'c', 'col2': 3}])
         self.assertEqual(g4, [{'col1': 'd', 'col2': 2}])
+
+
+class TestReadMe(unittest.TestCase):
+
+    def tearDown(self):
+        if Path('test/sample.db').is_file():
+            os.remove('test/sample.db')
+
+    def test_all_of_em(self):
+
+        t1 = [
+            {'col1': 'a', 'col2': 4},
+            {'col1': 'a', 'col2': 5},
+            {'col1': 'b', 'col2': 1},
+        ]
+
+        conn = Conn('test/sample.db')
+        conn['t1'] = t1
+
+        for r1, r2 in zip(conn['t1'].list(), conn['t1'].iter()):
+            self.assertEqual(r1, r2)
+
+        import pandas as pd
+
+        df = pd.DataFrame(conn['t1'].iter())
+        conn['t1_copy'] = df.to_dict('records')
+
+        self.assertEqual(conn['t1'].list(), conn['t1_copy'].list())
+
+        def add1(r):
+            if r['col2'] > 2:
+                r['col2'] += 1
+                return r
+
+        conn['t1_1'] = conn['t1'].map(add1)
+
+        self.assertEqual(conn['t1_1'].list()[0]['col2'], 5)
+
+        def sumup(rs):
+            r0 = rs[0]
+            r0['col2_sum'] = sum(r['col2'] for r in rs)
+            del r0['col2']
+            return r0
+
+        conn['t1_col2sum_groupby_col1'] = conn['t1'].by('col1')\
+            .map(sumup)
+
+        conn['t1_col2sum_groupby_col1_1'] = conn['t1']\
+            .order('col1', 'col2 desc').group('col1')\
+            .map(sumup)
+
+        self.assertEqual(conn['t1_col2sum_groupby_col1'].list(),
+                         conn['t1_col2sum_groupby_col1_1'].list())
+
+        conn['t1_double'] = conn['t1'].concat(conn['t1'])
+        self.assertEqual(len(conn['t1_double'].list()),
+                         2 * len(conn['t1'].list()))
+
+        conn['t1_double'] = conn['t1'].concat(t1)
+
+        self.assertEqual(len(conn['t1_double'].list()),
+                         2 * len(conn['t1'].list()))
+
+        conn['t1_1'] = conn['t1']\
+            .filter(lambda r: r['col2'] > 2)\
+            .update(
+                col2=lambda r: r['col2'] + 1,
+                col3=lambda r: r['col1'] + str(r['col2'])
+        )
+        self.assertEqual(conn['t1_1'].list(),
+                         [{'col1': 'a', 'col2': 5, 'col3': 'a5'},
+                          {'col1': 'a', 'col2': 6, 'col3': 'a6'}])
+
+        conn['t1_2'] = conn['t1'].by('col1').filter(lambda rs: len(rs) > 1)
+
+        self.assertEqual(conn['t1_2'].list(), [
+                         {'col1': 'a', 'col2': 4}, {'col1': 'a', 'col2': 5}])
+
+        conn['t1_1'] = conn['t1'].update(col3=lambda r: r['col2'] + 1)\
+            .deselect('col1', 'col2')
+
+        self.assertEqual(conn['t1_1'].list(),
+                         [{'col3': 5}, {'col3': 6}, {'col3': 2}])
+
+        conn['t1_1'] = conn['t1'].distinct('col1')
+        self.assertEqual(conn['t1_1'].list(), [
+                         {'col1': 'a', 'col2': 4}, {'col1': 'b', 'col2': 1}])
+
+        conn['t1_col2sum_groupby_col1'] = conn['t1'].by('col1')\
+            .fold(col2_sum=lambda rs: sum(r['col2'] for r in rs))\
+            .deselect('col2')
+
+        self.assertEqual(conn['t1_col2sum_groupby_col1'].list(), [
+                         {'col1': 'a', 'col2_sum': 9},
+                         {'col1': 'b', 'col2_sum': 1}])
+
+        conn['t2'] = [
+            {'col1': 'b', 'col3': -1},
+            {'col1': 'c', 'col3': 3},
+            {'col1': 'b', 'col3': None},
+        ]
+
+        conn['t1_col3'] = conn['t1'].by('col1')\
+            .full_join(conn['t2'].by('col1'))
+
+        self.assertEqual(conn['t1_col3'].list(),
+                         [{'col1': 'a', 'col2': 4, 'col3': None},
+                         {'col1': 'a', 'col2': 5, 'col3': None},
+                         {'col1': 'b', 'col2': 1, 'col3': -1},
+                         {'col1': 'b', 'col2': 1, 'col3': None},
+                          {'col1': 'c', 'col2': None, 'col3': 3}]
+                         )
